import asyncio
import sys

from tabulate import tabulate

from ...request import Request
from ...session import AsyncSession
from ..pretty import print_done, print_error, print_warn
from . import admin


@admin.group()
def license() -> None:
    """
    License administration commands.
    """


@license.command()
def show():
    """
    Show the license information (enterprise editions only).
    """
    async def _show_license():
        async with AsyncSession():
            rqst = Request('GET', '/license')
            async with rqst.fetch() as resp:
                data = await resp.json()
            if data['status'] == 'valid':
                print_done('Your Backend.AI lincese is valid.')
                print(tabulate([(k, v) for k, v in data['certificate'].items()]))
            else:
                print_warn('Your Backend.AI lincese is valid.')
    try:
        asyncio.run(_show_license())
    except Exception as e:
        print_error(e)
        sys.exit(1)



from . import admin  # noqa  # type: ignore
from . import config  # noqa  # type: ignore
from . import dotfile  # noqa  # type: ignore
from . import extensions  # noqa  # type: ignore
from . import main  # noqa  # type: ignore
from . import server_log  # noqa  # type: ignore
from . import session  # noqa  # type: ignore
from . import session_template  # noqa  # type: ignore
from . import vfolder  # noqa  # type: ignore
from . import app, logs, proxy, run  # noqa  # type: ignore



from ..main import main


@main.group()
def admin():
    """
    Administrative command set
    """


from . import (  # noqa
    agent,
    domain,
    etcd,
    group,
    image,
    keypair,
    license,
    manager,
    resource,
    resource_policy,
    scaling_group,
    session,
    storage,
    user,
    vfolder,
)



import getpass
import json
import sys
import warnings

import click

from .. import __version__
from ..config import get_config, local_state_path
from ..exceptions import BackendClientError
from ..session import Session
from .main import main
from .pretty import print_done, print_error, print_fail, print_warn


@main.command()
def config():
    '''
    Shows the current configuration.
    '''
    config = get_config()
    click.echo('API endpoint: {0} (mode: {1})'.format(
        click.style(str(config.endpoint), bold=True),
        click.style(str(config.endpoint_type), fg='cyan', bold=True)))
    click.echo('Client version: {0} (API: {1})'.format(
        click.style(__version__, bold=True),
        click.style(config.version, bold=True),
    ))
    if sys.stdout.isatty():
        click.echo('Server version: ...')
        click.echo('Negotiated API version: ...')
    else:
        with Session() as sess:
            try:
                versions = sess.System.get_versions()
            except BackendClientError:
                click.echo('Server version: (failed to fetch)')
            else:
                click.echo('Server version: {0} (API: {1})'.format(
                    versions.get('manager', 'pre-19.03'),
                    versions['version'],
                ))
            click.echo('Negotiated API version: {0}'.format(sess.api_version))
    nrows = 1
    if config.domain:
        click.echo('Domain name: "{0}"'.format(click.style(config.domain, bold=True)))
        nrows += 1
    if config.group:
        click.echo('Group name: "{0}"'.format(click.style(config.group, bold=True)))
        nrows += 1
    if config.is_anonymous:
        click.echo('Access key: (this is an anonymous session)')
        nrows += 1
    elif config.endpoint_type == 'docker':
        pass
    elif config.endpoint_type == 'session':
        if (local_state_path / 'cookie.dat').exists() and \
                (local_state_path / 'config.json').exists():
            sess_config = json.loads((local_state_path / 'config.json').read_text())
            click.echo('Username: "{0}"'.format(click.style(sess_config.get('username', ''), bold=True)))
            nrows += 1
    else:
        click.echo('Access key: "{0}"'.format(click.style(config.access_key, bold=True)))
        nrows += 1
        masked_skey = config.secret_key[:6] + ('*' * 24) + config.secret_key[-10:]
        click.echo('Secret key: "{0}"'.format(click.style(masked_skey, bold=True)))
        nrows += 1
    click.echo('Signature hash type: {0}'.format(
        click.style(config.hash_type, bold=True)))
    nrows += 1
    click.echo('Skip SSL certificate validation? {0}'.format(
        click.style(str(config.skip_sslcert_validation), bold=True)))
    nrows += 1
    if sys.stdout.isatty():
        sys.stdout.flush()
        with warnings.catch_warnings(record=True) as captured_warnings, Session() as sess:
            click.echo('\u001b[{0}A\u001b[2K'.format(nrows + 1), nl=False)
            try:
                versions = sess.System.get_versions()
            except BackendClientError:
                click.echo('Server version: {0}'.format(
                    click.style('(failed to fetch)', fg='red', bold=True),
                ))
            else:
                click.echo('Server version: {0} (API: {1})'.format(
                    click.style(versions.get('manager', 'pre-19.03'), bold=True),
                    click.style(versions['version'], bold=True),
                ))
            click.echo('\u001b[2K', nl=False)
            click.echo('Negotiated API version: {0}'.format(
                click.style('v{0[0]}.{0[1]}'.format(sess.api_version), bold=True),
            ))
            click.echo('\u001b[{0}B'.format(nrows), nl=False)
            sys.stdout.flush()
        for w in captured_warnings:
            warnings.showwarning(w.message, w.category, w.filename, w.lineno, w.line)


@main.command()
def login():
    '''
    Log-in to the console API proxy.
    It stores the current session cookie in the OS-default
    local application data location.
    '''
    user_id = input('User ID: ')
    password = getpass.getpass()

    config = get_config()
    if config.endpoint_type != 'session':
        print_warn('To use login, your endpoint type must be "session".')
        raise click.Abort()

    with Session() as session:
        try:
            result = session.Auth.login(user_id, password)
            if not result['authenticated']:
                print_fail('Login failed.')
                sys.exit(1)
            print_done('Login succeeded.')

            local_state_path.mkdir(parents=True, exist_ok=True)
            session.aiohttp_session.cookie_jar.update_cookies(result['cookies'])
            session.aiohttp_session.cookie_jar.save(local_state_path / 'cookie.dat')
            (local_state_path / 'config.json').write_text(json.dumps(result.get('config', {})))
        except Exception as e:
            print_error(e)


@main.command()
def logout():
    '''
    Log-out from the console API proxy and clears the local cookie data.
    '''
    config = get_config()
    if config.endpoint_type != 'session':
        print_warn('To use logout, your endpoint type must be "session".')
        raise click.Abort()

    with Session() as session:
        try:
            session.Auth.logout()
            print_done('Logout done.')
            try:
                (local_state_path / 'cookie.dat').unlink()
                (local_state_path / 'config.json').unlink()
            except (IOError, PermissionError):
                pass
        except Exception as e:
            print_error(e)


@main.command()
@click.argument('old_password', metavar='OLD_PASSWORD')
@click.argument('new_password', metavar='NEW_PASSWORD')
@click.argument('new_password2', metavar='NEW_PASSWORD2')
def update_password(old_password, new_password, new_password2):
    '''
    Update user's password.
    '''
    config = get_config()
    if config.endpoint_type != 'session':
        print_warn('To update password, your endpoint type must be "session".')
        raise click.Abort()

    with Session() as session:
        try:
            session.Auth.update_password(old_password, new_password, new_password2)
            print_done('Password updated.')
        except Exception as e:
            print_error(e)



"""
A support module to async mocks in Python versiosn prior to 3.8.
"""

import sys
from unittest import mock

if sys.version_info >= (3, 8, 0):
    # Since Python 3.8, AsyncMock is now part of the stdlib.
    # Python 3.8 also adds magic-mocking async iterators and async context managers.
    from unittest.mock import AsyncMock
else:
    from asynctest import CoroutineMock as AsyncMock  # type: ignore


class AsyncContextMock(mock.Mock):
    """
    Provides a mock that can be used:

        async with mock():
          ...

    Example:

        # In the test code:
        mock_obj = unittest.mock.Mock()
        mock_obj.fetch.return_value = AsyncContextMock(
            status=200,
            json=mock.AsyncMock(return_value={'hello': 'world'})
        )
        mocker.patch('mypkg.mymod.MyClass', return_value=mock_obj)

        # In the tested code:
        obj = mpkg.mymod.MyClass()
        async with obj.fetch() as resp:
            # resp.status is 200
            result = await resp.json()
            # result is {'hello': 'world'}
    """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


class AsyncContextMagicMock(mock.MagicMock):
    """
    Provides a magic mock that can be used:

        async with mock():
          ...
    """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass


class AsyncContextCoroutineMock(AsyncMock):
    """
    Provides a mock that can be used:

        async with (await mock(...)):
          ...

    Example:

        # In the test code:
        mock_obj = unittest.mock.AsyncMock()
        mock_obj.fetch.return_value = AsyncContextMock(
            status=200,
            json=mock.AsyncMock(return_value={'hello': 'world'})
        )
        mocker.patch('mypkg.mymod.MyClass', return_value=mock_obj)

        # In the tested code:
        obj = mpkg.mymod.MyClass()
        async with (await obj.fetch()) as resp:
            # resp.status is 200
            result = await resp.json()
            # result is {'hello': 'world'}
    """

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



import textwrap
from typing import Sequence

from ai.backend.client.output.fields import storage_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult
from ai.backend.client.pagination import generate_paginated_results
from ai.backend.client.session import api_session

from .base import BaseFunction, api_function

__all__ = (
    'Storage',
)

_default_list_fields = (
    storage_fields['id'],
    storage_fields['backend'],
    storage_fields['capabilities'],
)

_default_detail_fields = (
    storage_fields['id'],
    storage_fields['backend'],
    storage_fields['path'],
    storage_fields['fsprefix'],
    storage_fields['capabilities'],
    storage_fields['hardware_metadata'],
)


class Storage(BaseFunction):
    """
    Provides a shortcut of :func:`Admin.query()
    <ai.backend.client.admin.Admin.query>` that fetches various straoge volume
    information keyed by vfolder hosts.

    .. note::

      All methods in this function class require your API access key to
      have the *super-admin* privilege.
    """

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        status: str = 'ALIVE',
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str = None,
        order: str = None,
    ) -> PaginatedResult[dict]:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        return await generate_paginated_results(
            'storage_volume_list',
            {
                'filter': (filter, 'String'),
                'order': (order, 'String'),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    @classmethod
    async def detail(
        cls,
        vfolder_host: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        query = textwrap.dedent("""\
            query($vfolder_host: String!) {
                storage_volume(id: $vfolder_host) {$fields}
            }
        """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {'vfolder_host': vfolder_host}
        data = await api_session.get().Admin._query(query, variables)
        return data['storage_volume']



import enum
import functools
import sys
import textwrap
import traceback

from click import echo, style

from ..exceptions import BackendAPIError

__all__ = (
    'PrintStatus', 'print_pretty', 'print_info', 'print_wait',
    'print_done', 'print_warn', 'print_fail', 'print_error',
    'show_warning',
)


class PrintStatus(enum.Enum):
    NONE = 0
    WAITING = 1
    DONE = 2
    FAILED = 3
    WARNING = 4


def bold(text: str) -> str:
    '''
    Wraps the given text with bold enable/disable ANSI sequences.
    '''
    return (style(text, bold=True, reset=False) +
            style('', bold=False, reset=False))


def underline(text: str) -> str:
    return (style(text, underline=True, reset=False) +
            style('', underline=False, reset=False))


def inverse(text: str) -> str:
    return (style(text, reverse=True, reset=False) +
            style('', reverse=False, reset=False))


def italic(text: str) -> str:
    return '\x1b[3m' + text + '\x1b[23m'


def format_pretty(msg, status=PrintStatus.NONE, colored=True):
    if status == PrintStatus.NONE:
        indicator = style('\u2219', fg='bright_cyan', reset=False)
    elif status == PrintStatus.WAITING:
        indicator = style('\u22EF', fg='bright_yellow', reset=False)
    elif status == PrintStatus.DONE:
        indicator = style('\u2714', fg='bright_green', reset=False)
    elif status == PrintStatus.FAILED:
        indicator = style('\u2718', fg='bright_red', reset=False)
    elif status == PrintStatus.WARNING:
        indicator = style('\u2219', fg='yellow', reset=False)
    else:
        raise ValueError
    return style(indicator + textwrap.indent(msg, '  ')[1:], reset=True)


format_info = functools.partial(format_pretty, status=PrintStatus.NONE)
format_wait = functools.partial(format_pretty, status=PrintStatus.WAITING)
format_done = functools.partial(format_pretty, status=PrintStatus.DONE)
format_fail = functools.partial(format_pretty, status=PrintStatus.FAILED)
format_warn = functools.partial(format_pretty, status=PrintStatus.WARNING)


def print_pretty(msg, *, status=PrintStatus.NONE, file=None):
    if file is None:
        file = sys.stderr
    if status == PrintStatus.NONE:
        indicator = style('\u2219', fg='bright_cyan', reset=False)
    elif status == PrintStatus.WAITING:
        assert '\n' not in msg, 'Waiting message must be a single line.'
        indicator = style('\u22EF', fg='bright_yellow', reset=False)
    elif status == PrintStatus.DONE:
        indicator = style('\u2714', fg='bright_green', reset=False)
    elif status == PrintStatus.FAILED:
        indicator = style('\u2718', fg='bright_red', reset=False)
    elif status == PrintStatus.WARNING:
        indicator = style('\u2219', fg='yellow', reset=False)
    else:
        raise ValueError
    echo('\x1b[2K', nl=False, file=file)
    text = textwrap.indent(msg, '  ')
    text = style(indicator + text[1:], reset=True)
    echo('{0}\r'.format(text), nl=False, file=file)
    file.flush()
    if status != PrintStatus.WAITING:
        echo('', file=file)


print_info = functools.partial(print_pretty, status=PrintStatus.NONE)
print_wait = functools.partial(print_pretty, status=PrintStatus.WAITING)
print_done = functools.partial(print_pretty, status=PrintStatus.DONE)
print_fail = functools.partial(print_pretty, status=PrintStatus.FAILED)
print_warn = functools.partial(print_pretty, status=PrintStatus.WARNING)


def format_error(exc: Exception):
    if isinstance(exc, BackendAPIError):
        yield '{0}: {1} {2}\n'.format(exc.__class__.__name__,
                                      exc.status, exc.reason)
        yield '{0[title]}'.format(exc.data)
        if exc.data['type'].endswith('/too-many-sessions-matched'):
            matches = exc.data['data'].get('matches', [])
            if matches:
                yield "\nCandidates (up to 10 recent entries):\n"
            for item in matches:
                yield f"- {item['id']} ({item['name']}, {item['status']})\n"
        elif exc.data['type'].endswith('/session-already-exists'):
            existing_session_id = exc.data['data'].get('existingSessionId', None)
            if existing_session_id is not None:
                yield f"\n- Existing session ID: {existing_session_id}"
        elif exc.data['type'].endswith('/invalid-api-params'):
            general_error_msg = exc.data.get('msg', None)
            if general_error_msg is not None:
                yield f"\n- {general_error_msg}"
            per_field_errors = exc.data.get('data', {})
            if isinstance(per_field_errors, dict):
                for k, v in per_field_errors.items():
                    yield f"\n- \"{k}\": {v}"
            else:
                yield f"\n- {per_field_errors}"
        else:
            if exc.data['type'].endswith('/graphql-error'):
                yield "\n\u279c Message:\n"
                yield from (f"{err_item['message']}\n"
                            for err_item in exc.data.get('data', []))
            else:
                other_details = exc.data.get('msg', None)
                if other_details:
                    yield '\n\u279c Message: '
                    yield str(other_details)
                other_data = exc.data.get('data', None)
                if other_data:
                    yield '\n\u279c Data: '
                    yield repr(other_data)
        agent_details = exc.data.get('agent-details', None)
        if agent_details is not None:
            yield "\n\u279c This is an agent-side error. "
            yield "Check the agent status or ask the administrator for help."
            agent_exc = agent_details.get('exception', None)
            if agent_exc is not None:
                yield '\n\u279c ' + str(agent_exc)
            desc = agent_details.get('title', None)
            if desc is not None:
                yield '\n\u279c ' + str(desc)
        content = exc.data.get('content', None)
        if content:
            yield "\n" + content
    else:
        args = exc.args if exc.args else ['']
        yield f"{exc.__class__.__name__}: {args[0]}\n"
        yield "\n".join(map(str, args[1:]))
        yield ("*** Traceback ***\n" +
               "".join(traceback.format_tb(exc.__traceback__)).strip())


def print_error(exc: Exception, *, file=None):
    if file is None:
        file = sys.stderr
    indicator = style('\u2718', fg='bright_red', reset=False)
    if file.isatty():
        echo('\x1b[2K', nl=False, file=file)
    text = ''.join(format_error(exc))
    text = textwrap.indent(text, '  ')
    text = style(indicator + text[1:], reset=True)
    echo('{0}\r'.format(text), nl=False, file=file)
    echo('', file=file)
    file.flush()


def show_warning(message, category, filename, lineno, file=None, line=None):
    echo('{0}: {1}'.format(
        style(str(category.__name__), fg='yellow', bold=True),
        style(str(message), fg='yellow'),
    ), file=file)



import sys

import click

from ai.backend.client.func.scaling_group import (
    _default_detail_fields,
    _default_list_fields,
)
from ai.backend.client.output.fields import scaling_group_fields
from ai.backend.client.session import Session

from ..extensions import pass_ctx_obj
from ..params import JSONParamType
from ..types import CLIContext
from . import admin


@admin.group()
def scaling_group() -> None:
    """
    Scaling group (resource group) administration commands.
    """


@scaling_group.command()
@pass_ctx_obj
@click.argument('group', type=str, metavar='GROUP_NAME')
def get_available(ctx: CLIContext, group: str) -> None:
    with Session() as session:
        try:
            items = session.ScalingGroup.list_available(group)
            ctx.output.print_list(items, [scaling_group_fields['name']])
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@scaling_group.command()
@pass_ctx_obj
@click.argument('name', type=str)
def info(ctx: CLIContext, name: str) -> None:
    """
    Show the information about the given scaling group.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            item = session.ScalingGroup.detail(name=name)
            ctx.output.print_item(item, _default_detail_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@scaling_group.command()
@pass_ctx_obj
def list(ctx: CLIContext) -> None:
    """
    List and manage scaling groups.
    (superadmin privilege required)
    """
    with Session() as session:
        try:
            items = session.ScalingGroup.list()
            ctx.output.print_list(items, _default_list_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@scaling_group.command()
@pass_ctx_obj
@click.argument('name', type=str, metavar='NAME')
@click.option('-d', '--description', type=str, default='',
              help='Description of new scaling group')
@click.option('-i', '--inactive', is_flag=True,
              help='New scaling group will be inactive.')
@click.option('--driver', type=str, default='static',
              help='Set driver.')
@click.option('--driver-opts', type=JSONParamType(), default='{}',
              help='Set driver options as a JSON string.')
@click.option('--scheduler', type=str, default='fifo',
              help='Set scheduler.')
@click.option('--scheduler-opts', type=JSONParamType(), default='{}',
              help='Set scheduler options as a JSON string.')
def add(ctx: CLIContext, name, description, inactive,
        driver, driver_opts, scheduler, scheduler_opts):
    """
    Add a new scaling group.

    NAME: Name of new scaling group.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.create(
                name,
                description=description,
                is_active=not inactive,
                driver=driver,
                driver_opts=driver_opts,
                scheduler=scheduler,
                scheduler_opts=scheduler_opts,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name='scaling_group',
                action_name='add',
            )
            sys.exit(1)
        if not data['ok']:
            ctx.output.print_mutation_error(
                msg=data['msg'],
                item_name='scaling_group',
                action_name='add',
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            item_name='scaling_group',
        )


@scaling_group.command()
@pass_ctx_obj
@click.argument('name', type=str, metavar='NAME')
@click.option('-d', '--description', type=str, default='',
              help='Description of new scaling group')
@click.option('-i', '--inactive', is_flag=True,
              help='New scaling group will be inactive.')
@click.option('--driver', type=str, default='static',
              help='Set driver.')
@click.option('--driver-opts', type=JSONParamType(), default=None,
              help='Set driver options as a JSON string.')
@click.option('--scheduler', type=str, default='fifo',
              help='Set scheduler.')
@click.option('--scheduler-opts', type=JSONParamType(), default=None,
              help='Set scheduler options as a JSON string.')
def update(ctx: CLIContext, name, description, inactive,
           driver, driver_opts, scheduler, scheduler_opts):
    """
    Update existing scaling group.

    NAME: Name of new scaling group.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.update(
                name,
                description=description,
                is_active=not inactive,
                driver=driver,
                driver_opts=driver_opts,
                scheduler=scheduler,
                scheduler_opts=scheduler_opts,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name='scaling_group',
                action_name='update',
            )
            sys.exit(1)
        if not data['ok']:
            ctx.output.print_mutation_error(
                msg=data['msg'],
                item_name='scaling_group',
                action_name='update',
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                'name': name,
            },
        )


@scaling_group.command()
@pass_ctx_obj
@click.argument('name', type=str, metavar='NAME')
def delete(ctx: CLIContext, name):
    """
    Delete an existing scaling group.

    NAME: Name of a scaling group to delete.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.delete(name)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name='scaling_group',
                action_name='deletion',
            )
            sys.exit(1)
        if not data['ok']:
            ctx.output.print_mutation_error(
                msg=data['msg'],
                item_name='scaling_group',
                action_name='deletion',
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                'name': name,
            },
        )


@scaling_group.command()
@pass_ctx_obj
@click.argument('scaling_group', type=str, metavar='SCALING_GROUP')
@click.argument('domain', type=str, metavar='DOMAIN')
def associate_scaling_group(ctx: CLIContext, scaling_group, domain):
    """
    Associate a domain with a scaling_group.

    \b
    SCALING_GROUP: The name of a scaling group.
    DOMAIN: The name of a domain.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.associate_domain(scaling_group, domain)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name='scaling_group',
                action_name='scaling_group_association',
            )
            sys.exit(1)
        if not data['ok']:
            ctx.output.print_mutation_error(
                msg=data['msg'],
                item_name='scaling_group',
                action_name='scaling_group_association',
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                'detail_msg': 'Scaling group {} is assocatiated with domain {}.'
                .format(scaling_group, domain),
            },
        )


@scaling_group.command()
@pass_ctx_obj
@click.argument('scaling_group', type=str, metavar='SCALING_GROUP')
@click.argument('domain', type=str, metavar='DOMAIN')
def dissociate_scaling_group(ctx: CLIContext, scaling_group, domain):
    """
    Dissociate a domain from a scaling_group.

    \b
    SCALING_GROUP: The name of a scaling group.
    DOMAIN: The name of a domain.
    """
    with Session() as session:
        try:
            data = session.ScalingGroup.dissociate_domain(scaling_group, domain)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name='scaling_group',
                action_name='scaling_group_dissociation',
            )
            sys.exit(1)
        if not data['ok']:
            ctx.output.print_mutation_error(
                msg=data['msg'],
                item_name='scaling_group',
                action_name='scaling_group_dissociation',
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            item_name='scaling_group',
            extra_info={
                'detail_msg': 'Scaling group {} is dissociated from domain {}.'
                .format(scaling_group, domain),
            },
        )



from typing import Any, List, Mapping

from ..request import Request
from ..session import api_session
from .base import BaseFunction, api_function

__all__ = (
    'SessionTemplate',
)


class SessionTemplate(BaseFunction):

    @api_function
    @classmethod
    async def create(cls,
                     template: str,
                     domain_name: str = None,
                     group_name: str = None,
                     owner_access_key: str = None,
                     ) -> 'SessionTemplate':
        rqst = Request('POST', '/template/session')
        if domain_name is None:
            # Even if config.domain is None, it can be guessed in the manager by user information.
            domain_name = api_session.get().config.domain
        if group_name is None:
            group_name = api_session.get().config.group
        body = {
            'payload': template,
            'group_name': group_name,
            'domain_name': domain_name,
            'owner_access_key': owner_access_key,
        }
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            response = await resp.json()
            return cls(response['id'], owner_access_key=owner_access_key)

    @api_function
    @classmethod
    async def list_templates(cls, list_all: bool = False) -> List[Mapping[str, str]]:
        rqst = Request('GET', '/template/session')
        rqst.set_json({'all': list_all})
        async with rqst.fetch() as resp:
            return await resp.json()

    def __init__(self, template_id: str, owner_access_key: str = None):
        self.template_id = template_id
        self.owner_access_key = owner_access_key

    @api_function
    async def get(self, body_format: str = 'yaml') -> str:
        params = {'format': body_format}
        if self.owner_access_key:
            params['owner_access_key'] = self.owner_access_key
        rqst = Request('GET', f'/template/session/{self.template_id}',
                       params=params)
        async with rqst.fetch() as resp:
            data = await resp.text()
        return data

    @api_function
    async def put(self, template: str) -> Any:
        body = {
            'payload': template,
        }
        if self.owner_access_key:
            body['owner_access_key'] = self.owner_access_key
        rqst = Request('PUT', f'/template/session/{self.template_id}')
        rqst.set_json(body)
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete(self) -> Any:
        params = {}
        if self.owner_access_key:
            params['owner_access_key'] = self.owner_access_key
        rqst = Request('DELETE', f'/template/session/{self.template_id}',
                       params=params)
        async with rqst.fetch() as resp:
            return await resp.json()



from typing import Final

KERNEL_CANCELLED: Final[str] = 'kernel_cancelled'
KERNEL_CREATING: Final[str] = 'kernel_creating'
KERNEL_STARTED: Final[str] = 'kernel_started'
KERNEL_TERMINATED: Final[str] = 'kernel_terminated'
KERNEL_TERMINATING: Final[str] = 'kernel_terminating'
SESSION_FAILURE: Final[str] = 'session_failure'
SESSION_STARTED: Final[str] = 'session_started'
SESSION_SUCCESS: Final[str] = 'session_success'
SESSION_TERMINATED: Final[str] = 'session_terminated'



import sys

import click
from tabulate import tabulate

from ..session import Session
from .main import main
from .pretty import print_error, print_info, print_warn


@main.group()
def dotfile():
    '''Provides dotfile operations.'''


@dotfile.command()
@click.argument('path', metavar='PATH')
@click.option('--perm', 'permission',
             help='Linux permission represented in octal number (e.g. 755) '
                  'Defaults to 755 if not specified')
@click.option('-f', '--file', 'dotfile_path',
              help='Path to dotfile to upload. '
                   'If not specified, client will try to read file from STDIN. ')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session of user dotfiles.')
@click.option('-d', '--domain', 'domain', metavar='DOMAIN',
              help='Specify the domain name of domain dotfiles.')
@click.option('-g', '--group', metavar='GROUP',
              help='Sepcify the group name or id of group dotfiles. '
                   '(If group name is provided, domain name must be specified with option -d)')
def create(path, permission, dotfile_path, owner_access_key, domain, group):
    '''
    Store dotfile to Backend.AI Manager.
    Dotfiles will be automatically loaded when creating kernels.

    PATH: Where dotfiles will be created when starting kernel
    '''

    if dotfile_path:
        with open(dotfile_path, 'r') as fr:
            body = fr.read()
    else:
        body = ''
        for line in sys.stdin:
            body += (line + '\n')
    with Session() as session:
        try:
            if not permission:
                permission = '755'
            dotfile_ = session.Dotfile.create(body, path, permission,
                                              owner_access_key=owner_access_key,
                                              domain=domain, group=group)
            print_info(f'Dotfile {dotfile_.path} created and ready')
        except Exception as e:
            print_error(e)
            sys.exit(1)


@dotfile.command()
@click.argument('path', metavar='PATH')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session of user dotfiles.')
@click.option('-d', '--domain', 'domain', metavar='DOMAIN',
              help='Specify the domain name of domain dotfiles.')
@click.option('-g', '--group', metavar='GROUP',
              help='Sepcify the group name or id of group dotfiles. '
                   '(If group name is provided, domain name must be specified with option -d)')
def get(path, owner_access_key, domain, group):
    '''
    Print dotfile content.
    '''
    with Session() as session:
        try:
            dotfile_ = session.Dotfile(path, owner_access_key=owner_access_key,
                                       domain=domain, group=group)
            body = dotfile_.get()
            print(body['data'])
        except Exception as e:
            print_error(e)
            sys.exit(1)


@dotfile.command()
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session of user dotfiles.')
@click.option('-d', '--domain', 'domain', metavar='DOMAIN',
              help='Specify the domain name of domain dotfiles.')
@click.option('-g', '--group', metavar='GROUP',
              help='Sepcify the group name or id of group dotfiles. '
                   '(If group name is provided, domain name must be specified with option -d)')
def list(owner_access_key, domain, group):
    '''
    List availabe user/domain/group dotfiles.
    '''
    fields = [
        ('Path', 'path', None),
        ('Data', 'data', lambda v: v[:30].splitlines()[0]),
        ('Permission', 'permission', None),
    ]
    with Session() as session:
        try:
            resp = session.Dotfile.list_dotfiles(owner_access_key=owner_access_key,
                                                 domain=domain, group=group)
            if not resp:
                print('There is no dotfiles created yet.')
                return
            rows = (
                tuple(
                    item[key] if transform is None else transform(item[key])
                    for _, key, transform in fields
                )
                for item in resp
            )
            hdrs = (display_name for display_name, _, _ in fields)
            print(tabulate(rows, hdrs))
        except Exception as e:
            print_error(e)
            sys.exit(1)


@dotfile.command()
@click.argument('path', metavar='PATH')
@click.option('--perm', 'permission',
             help='Linux permission represented in octal number (e.g. 755) '
                  'Defaults to 755 if not specified')
@click.option('-f', '--file', 'dotfile_path',
              help='Path to dotfile to upload. '
                   'If not specified, client will try to read file from STDIN. ')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session of user dotfiles.')
@click.option('-d', '--domain', 'domain', metavar='DOMAIN',
              help='Specify the domain name of domain dotfiles.')
@click.option('-g', '--group', metavar='GROUP',
              help='Sepcify the group name or id of group dotfiles. '
                   '(If group name is provided, domain name must be specified with option -d)')
def update(path, permission, dotfile_path, owner_access_key, domain, group):
    '''
    Update dotfile stored in Backend.AI Manager.
    '''

    if dotfile_path:
        with open(dotfile_path, 'r') as fr:
            body = fr.read()
    else:
        body = ''
        for line in sys.stdin:
            body += (line + '\n')
    with Session() as session:
        try:
            if not permission:
                permission = '755'
            dotfile_ = session.Dotfile(path, owner_access_key=owner_access_key,
                                       domain=domain, group=group)
            dotfile_.update(body, permission)
            print_info(f'Dotfile {dotfile_.path} updated')
        except Exception as e:
            print_error(e)
            sys.exit(1)


@dotfile.command()
@click.argument('path', metavar='PATH')
@click.option('-f', '--force', type=bool, is_flag=True,
              help='Delete dotfile without confirmation.')
@click.option('-o', '--owner', '--owner-access-key', 'owner_access_key', metavar='ACCESS_KEY',
              help='Specify the owner of the target session of user dotfiles.')
@click.option('-d', '--domain', 'domain', metavar='DOMAIN',
              help='Specify the domain name of domain dotfiles.')
@click.option('-g', '--group', metavar='GROUP',
              help='Sepcify the group name or id of group dotfiles. '
                   '(If group name is provided, domain name must be specified with option -d)')
def delete(path, force, owner_access_key, domain, group):
    '''
    Delete dotfile from Backend.AI Manager.
    '''
    with Session() as session:
        dotfile_ = session.Dotfile(path, owner_access_key=owner_access_key,
                                   domain=domain, group=group)
        if not force:
            print_warn('Are you sure? (y/[n])')
            result = input()
            if result.strip() != 'y':
                print_info('Aborting.')
                exit()
        try:
            dotfile_.delete()
            print_info(f'Dotfile {dotfile_.path} deleted')
        except Exception as e:
            print_error(e)
            sys.exit(1)



from typing import Any, Dict, Sequence, Union

from ai.backend.client.output.fields import keypair_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult
from ai.backend.client.pagination import generate_paginated_results
from ai.backend.client.session import api_session

from .base import BaseFunction, api_function

__all__ = (
    'KeyPair',
)

_default_list_fields = (
    keypair_fields['user_id'],
    keypair_fields['access_key'],
    keypair_fields['secret_key'],
    keypair_fields['is_active'],
    keypair_fields['is_admin'],
    keypair_fields['created_at'],
)

_default_detail_fields = (
    keypair_fields['user_id'],
    keypair_fields['access_key'],
    keypair_fields['secret_key'],
    keypair_fields['is_active'],
    keypair_fields['is_admin'],
)

_default_result_fields = (
    keypair_fields['access_key'],
    keypair_fields['secret_key'],
)


class KeyPair(BaseFunction):
    """
    Provides interactions with keypairs.
    """

    def __init__(self, access_key: str):
        self.access_key = access_key

    @api_function
    @classmethod
    async def create(
        cls,
        user_id: Union[int, str],
        is_active: bool = True,
        is_admin: bool = False,
        resource_policy: str = None,
        rate_limit: int = None,
        fields: Sequence[FieldSpec] = _default_result_fields,
    ) -> dict:
        """
        Creates a new keypair with the given options.
        You need an admin privilege for this operation.
        """
        uid_type = 'Int!' if isinstance(user_id, int) else 'String!'
        q = 'mutation($user_id: {0}, $input: KeyPairInput!) {{'.format(uid_type) + \
            '  create_keypair(user_id: $user_id, props: $input) {' \
            '    ok msg keypair { $fields }' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {
            'user_id': user_id,
            'input': {
                'is_active': is_active,
                'is_admin': is_admin,
                'resource_policy': resource_policy,
                'rate_limit': rate_limit,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['create_keypair']

    @api_function
    @classmethod
    async def update(
        cls,
        access_key: str,
        is_active: bool = None,
        is_admin: bool = None,
        resource_policy: str = None,
        rate_limit: int = None,
    ) -> dict:
        """
        Creates a new keypair with the given options.
        You need an admin privilege for this operation.
        """
        q = 'mutation($access_key: String!, $input: ModifyKeyPairInput!) {' + \
            '  modify_keypair(access_key: $access_key, props: $input) {' \
            '    ok msg' \
            '  }' \
            '}'
        variables = {
            'access_key': access_key,
            'input': {
                'is_active': is_active,
                'is_admin': is_admin,
                'resource_policy': resource_policy,
                'rate_limit': rate_limit,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['modify_keypair']

    @api_function
    @classmethod
    async def delete(cls, access_key: str):
        """
        Deletes an existing keypair with given ACCESSKEY.
        """
        q = 'mutation($access_key: String!) {' \
            '  delete_keypair(access_key: $access_key) {' \
            '    ok msg' \
            '  }' \
            '}'
        variables = {
            'access_key': access_key,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['delete_keypair']

    @api_function
    @classmethod
    async def list(
        cls,
        user_id: Union[int, str] = None,
        is_active: bool = None,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        if user_id is None:
            q = 'query($is_active: Boolean) {' \
                '  keypairs(is_active: $is_active) {' \
                '    $fields' \
                '  }' \
                '}'
        else:
            uid_type = 'Int!' if isinstance(user_id, int) else 'String!'
            q = 'query($email: {0}, $is_active: Boolean) {{'.format(uid_type) + \
                '  keypairs(email: $email, is_active: $is_active) {' \
                '    $fields' \
                '  }' \
                '}'
        q = q.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables: Dict[str, Any] = {
            'is_active': is_active,
        }
        if user_id is not None:
            variables['email'] = user_id
        data = await api_session.get().Admin._query(q, variables)
        return data['keypairs']

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        is_active: bool = None,
        domain_name: str = None,
        *,
        user_id: str = None,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str = None,
        order: str = None,
    ) -> PaginatedResult[dict]:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        variables = {
            'is_active': (is_active, 'Boolean'),
            'domain_name': (domain_name, 'String'),
            'filter': (filter, 'String'),
            'order': (order, 'String'),
        }
        if user_id is not None:
            variables['email'] = (user_id, 'String')
        return await generate_paginated_results(
            'keypair_list',
            variables,
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    async def info(self, fields: Sequence[FieldSpec] = _default_detail_fields) -> dict:
        """
        Returns the keypair's information such as resource limits.

        :param fields: Additional per-agent query fields to fetch.

        .. versionadded:: 18.12
        """
        q = 'query {' \
            '  keypair {' \
            '    $fields' \
            '  }' \
            '}'
        q = q.replace('$fields', ' '.join(f.field_ref for f in fields))
        data = await api_session.get().Admin._query(q)
        return data['keypair']

    @api_function
    @classmethod
    async def activate(cls, access_key: str) -> dict:
        """
        Activates this keypair.
        You need an admin privilege for this operation.
        """
        q = 'mutation($access_key: String!, $input: ModifyKeyPairInput!) {' + \
            '  modify_keypair(access_key: $access_key, props: $input) {' \
            '    ok msg' \
            '  }' \
            '}'
        variables = {
            'access_key': access_key,
            'input': {
                'is_active': True,
                'is_admin': None,
                'resource_policy': None,
                'rate_limit': None,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['modify_keypair']

    @api_function
    @classmethod
    async def deactivate(cls, access_key: str) -> dict:
        """
        Deactivates this keypair.
        Deactivated keypairs cannot make any API requests
        unless activated again by an administrator.
        You need an admin privilege for this operation.
        """
        q = 'mutation($access_key: String!, $input: ModifyKeyPairInput!) {' + \
            '  modify_keypair(access_key: $access_key, props: $input) {' \
            '    ok msg' \
            '  }' \
            '}'
        variables = {
            'access_key': access_key,
            'input': {
                'is_active': False,
                'is_admin': None,
                'resource_policy': None,
                'rate_limit': None,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data['modify_keypair']



import sys

import click

from ai.backend.cli.interaction import ask_yn
from ai.backend.client.func.keypair_resource_policy import (
    _default_detail_fields,
    _default_list_fields,
)
from ai.backend.client.session import Session

from ..extensions import pass_ctx_obj
from ..pretty import print_info
from ..types import CLIContext

# from ai.backend.client.output.fields import keypair_resource_policy_fields
from . import admin


@admin.group()
def keypair_resource_policy() -> None:
    """
    KeyPair resource policy administration commands.
    """


@keypair_resource_policy.command()
@pass_ctx_obj
@click.argument('name', type=str)
def info(ctx: CLIContext, name: str) -> None:
    """
    Show details about a keypair resource policy. When `name` option is omitted, the
    resource policy for the current access_key will be returned.
    """
    with Session() as session:
        try:
            rp = session.KeypairResourcePolicy(session.config.access_key)
            item = rp.info(name)
            ctx.output.print_item(item, _default_detail_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@keypair_resource_policy.command()
@pass_ctx_obj
def list(ctx):
    """
    List and manage keypair resource policies.
    (admin privilege required)
    """
    with Session() as session:
        try:
            items = session.KeypairResourcePolicy.list()
            ctx.output.print_list(items, _default_list_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(1)


@keypair_resource_policy.command()
@pass_ctx_obj
@click.argument('name', type=str, default=None, metavar='NAME')
@click.option('--default-for-unspecified', type=str, default='UNLIMITED',
              help='Default behavior for unspecified resources: '
                   'LIMITED, UNLIMITED')
@click.option('--total-resource-slots', type=str, default='{}',
              help='Set total resource slots.')
@click.option('--max-concurrent-sessions', type=int, default=30,
              help='Number of maximum concurrent sessions.')
@click.option('--max-containers-per-session', type=int, default=1,
              help='Number of maximum containers per session.')
@click.option('--max-vfolder-count', type=int, default=10,
              help='Number of maximum virtual folders allowed.')
@click.option('--max-vfolder-size', type=int, default=0,
              help='Maximum virtual folder size (future plan).')
@click.option('--idle-timeout', type=int, default=1800,
              help='The maximum period of time allowed for kernels to wait '
                   'further requests.')
# @click.option('--allowed-vfolder-hosts', type=click.Tuple(str), default=['local'],
#               help='Locations to create virtual folders.')
@click.option('--allowed-vfolder-hosts', default=['local'],
              help='Locations to create virtual folders.')
def add(ctx: CLIContext, name, default_for_unspecified, total_resource_slots, max_concurrent_sessions,
        max_containers_per_session, max_vfolder_count, max_vfolder_size,
        idle_timeout, allowed_vfolder_hosts):
    """
    Add a new keypair resource policy.

    NAME: NAME of a new keypair resource policy.
    """
    with Session() as session:
        try:
            data = session.KeypairResourcePolicy.create(
                name,
                default_for_unspecified=default_for_unspecified,
                total_resource_slots=total_resource_slots,
                max_concurrent_sessions=max_concurrent_sessions,
                max_containers_per_session=max_containers_per_session,
                max_vfolder_count=max_vfolder_count,
                max_vfolder_size=max_vfolder_size,
                idle_timeout=idle_timeout,
                allowed_vfolder_hosts=allowed_vfolder_hosts,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name='resource_policy',
                action_name='add',
            )
            sys.exit(1)
        if not data['ok']:
            ctx.output.print_mutation_error(
                msg=data['msg'],
                item_name='resource_policy',
                action_name='add',
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            item_name='resource_policy',
        )


@keypair_resource_policy.command()
@pass_ctx_obj
@click.argument('name', type=str, default=None, metavar='NAME')
@click.option('--default-for-unspecified', type=str,
              help='Default behavior for unspecified resources: '
                   'LIMITED, UNLIMITED')
@click.option('--total-resource-slots', type=str,
              help='Set total resource slots.')
@click.option('--max-concurrent-sessions', type=int,
              help='Number of maximum concurrent sessions.')
@click.option('--max-containers-per-session', type=int,
              help='Number of maximum containers per session.')
@click.option('--max-vfolder-count', type=int,
              help='Number of maximum virtual folders allowed.')
@click.option('--max-vfolder-size', type=int,
              help='Maximum virtual folder size (future plan).')
@click.option('--idle-timeout', type=int,
              help='The maximum period of time allowed for kernels to wait '
                   'further requests.')
@click.option('--allowed-vfolder-hosts', help='Locations to create virtual folders.')
def update(ctx: CLIContext, name, default_for_unspecified, total_resource_slots,
           max_concurrent_sessions, max_containers_per_session, max_vfolder_count,
           max_vfolder_size, idle_timeout, allowed_vfolder_hosts):
    """
    Update an existing keypair resource policy.

    NAME: NAME of a keypair resource policy to update.
    """
    with Session() as session:
        try:
            data = session.KeypairResourcePolicy.update(
                name,
                default_for_unspecified=default_for_unspecified,
                total_resource_slots=total_resource_slots,
                max_concurrent_sessions=max_concurrent_sessions,
                max_containers_per_session=max_containers_per_session,
                max_vfolder_count=max_vfolder_count,
                max_vfolder_size=max_vfolder_size,
                idle_timeout=idle_timeout,
                allowed_vfolder_hosts=allowed_vfolder_hosts,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name='resource_policy',
                action_name='update',
            )
            sys.exit(1)
        if not data['ok']:
            ctx.output.print_mutation_error(
                msg=data['msg'],
                item_name='resource_policy',
                action_name='update',
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                'name': name,
            },
        )


@keypair_resource_policy.command()
@pass_ctx_obj
@click.argument('name', type=str, default=None, metavar='NAME')
def delete(ctx: CLIContext, name):
    """
    Delete a keypair resource policy.

    NAME: NAME of a keypair resource policy to delete.
    """
    with Session() as session:
        if not ask_yn():
            print_info('Cancelled.')
            sys.exit(1)
        try:
            data = session.KeypairResourcePolicy.delete(name)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name='resource_policy',
                action_name='deletion',
            )
            sys.exit(1)
        if not data['ok']:
            ctx.output.print_mutation_error(
                msg=data['msg'],
                item_name='resource_policy',
                action_name='deletion',
            )
            sys.exit(1)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                'name': name,
            },
        )



from typing import Any, Mapping, Optional

from ..exceptions import BackendAPIError
from ..request import Request
from ..session import api_session
from .base import BaseFunction, api_function

__all__ = (
    'Admin',
)


class Admin(BaseFunction):
    """
    Provides the function interface for making admin GrapQL queries.

    .. note::

      Depending on the privilege of your API access key, you may or may not
      have access to querying/mutating server-side resources of other
      users.
    """

    @api_function
    @classmethod
    async def query(
        cls,
        query: str,
        variables: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        """
        Sends the GraphQL query and returns the response.

        :param query: The GraphQL query string.
        :param variables: An optional key-value dictionary
            to fill the interpolated template variables
            in the query.

        :returns: The object parsed from the response JSON string.
        """
        return await cls._query(query, variables)

    @classmethod
    async def _query(
        cls,
        query: str,
        variables: Optional[Mapping[str, Any]] = None,
    ) -> Any:
        """
        Internal async implementation of the query() method,
        which may be reused by other functional APIs to make GQL requests.
        """
        gql_query = {
            'query': query,
            'variables': variables if variables else {},
        }
        if api_session.get().api_version >= (6, '20210815'):
            rqst = Request('POST', '/admin/gql')
            rqst.set_json(gql_query)
            async with rqst.fetch() as resp:
                response = await resp.json()
                errors = response.get("errors", [])
                if errors:
                    raise BackendAPIError(400, reason="Bad request", data={
                        'type': 'https://api.backend.ai/probs/graphql-error',
                        'title': 'GraphQL-generated error',
                        'data': errors,
                    })
                else:
                    return response["data"]
        else:
            rqst = Request('POST', '/admin/graphql')
            rqst.set_json(gql_query)
            async with rqst.fetch() as resp:
                return await resp.json()



from __future__ import annotations

import sys
from typing import Any, Callable, Mapping, Optional, Sequence

from tabulate import tabulate

from ai.backend.client.cli.pagination import (
    echo_via_pager,
    get_preferred_page_size,
    tabulate_items,
)
from ai.backend.client.cli.pretty import print_error, print_fail

from .types import BaseOutputHandler, FieldSpec, PaginatedResult


class NoItems(Exception):
    pass


class ConsoleOutputHandler(BaseOutputHandler):

    def print_item(
        self,
        item: Mapping[str, Any] | None,
        fields: Sequence[FieldSpec],
    ) -> None:
        if item is None:
            print_fail("No matching entry found.")
            return
        field_map = {f.field_name: f for f in fields}
        print(tabulate(
            [
                (
                    field_map[k].humanized_name,
                    field_map[k].formatter.format_console(v, field_map[k]),
                )
                for k, v in item.items()
            ],
            headers=('Field', 'Value'),
        ))

    def print_items(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
    ) -> None:
        field_map = {f.field_name: f for f in fields}
        for idx, item in enumerate(items):
            if idx > 0:
                print("-" * 20)
            print(tabulate(
                [
                    (
                        field_map[k].humanized_name,
                        field_map[k].formatter.format_console(v, field_map[k]),
                    )
                    for k, v in item.items()
                ],
                headers=('Field', 'Value'),
            ))

    def print_list(
        self,
        items: Sequence[Mapping[str, Any]],
        fields: Sequence[FieldSpec],
        *,
        is_scalar: bool = False,
    ) -> None:
        if is_scalar:
            assert len(fields) == 1
        if sys.stdout.isatty():

            def infinite_fetch():
                current_offset = 0
                page_size = get_preferred_page_size()
                while True:
                    if len(items) == 0:
                        raise NoItems
                    if is_scalar:
                        yield from map(
                            lambda v: {fields[0].field_name: v},
                            items[current_offset:current_offset + page_size],
                        )
                    else:
                        yield from items[current_offset:current_offset + page_size]
                    current_offset += page_size
                    if current_offset >= len(items):
                        break

            try:
                echo_via_pager(
                    tabulate_items(
                        infinite_fetch(),
                        fields,
                    ),
                )
            except NoItems:
                print("No matching items.")
        else:
            if is_scalar:
                for line in tabulate_items(
                    map(lambda v: {fields[0].field_name: v}, items),  # type: ignore
                    fields,
                ):
                    print(line, end="")
            else:
                for line in tabulate_items(
                    items,  # type: ignore
                    fields,
                ):
                    print(line, end="")

    def print_paginated_list(
        self,
        fetch_func: Callable[[int, int], PaginatedResult],
        initial_page_offset: int,
        page_size: int = None,
    ) -> None:
        if sys.stdout.isatty() and page_size is None:
            page_size = get_preferred_page_size()
            fields: Sequence[FieldSpec] = []

            def infinite_fetch():
                nonlocal fields
                current_offset = initial_page_offset
                while True:
                    result = fetch_func(current_offset, page_size)
                    if result.total_count == 0:
                        raise NoItems
                    current_offset += len(result.items)
                    if not fields:
                        fields.extend(result.fields)
                    yield from result.items
                    if current_offset >= result.total_count:
                        break

            try:
                echo_via_pager(
                    tabulate_items(
                        infinite_fetch(),
                        fields,
                    ),
                )
            except NoItems:
                print("No matching items.")
        else:
            page_size = page_size or 20
            result = fetch_func(initial_page_offset, page_size)
            for line in tabulate_items(
                result.items,  # type: ignore
                result.fields,
            ):
                print(line, end="")

    def print_mutation_result(
        self,
        item: Mapping[str, Any],
        item_name: Optional[str] = None,
        action_name: Optional[str] = None,
        extra_info: Mapping = {},
    ) -> None:
        t = [
            ['ok', item['ok']],
            ['msg', item['msg']],
            *[(k, v) for k, v in extra_info.items()],
        ]
        if action_name is not None:
            t += [['Action', action_name]]
        if item_name is not None:
            t += [(k, v) for k, v in item[item_name].items()]
        print(tabulate(
            t, headers=('Field', 'Value'),
        ))

    def print_mutation_error(
        self,
        error: Optional[Exception] = None,
        msg: str = 'Failed',
        item_name: Optional[str] = None,
        action_name: Optional[str] = None,
        extra_info: Mapping = {},
    ) -> None:
        t = [
            ['Message', msg],
        ]
        if item_name is not None:
            t += [['Item', item_name]]
        if action_name is not None:
            t += [['Action', action_name]]
        print(tabulate(
            t, headers=('Field', 'Value'),
        ))
        if error is not None:
            print_error(error)

    def print_error(
        self,
        error: Exception,
    ) -> None:
        print_error(error)

    def print_fail(
        self,
        message: str,
    ) -> None:
        print_fail(message)



from __future__ import annotations

import textwrap
from typing import Any, Dict, Final, Sequence, Tuple, TypeVar

from .exceptions import BackendAPIVersionError
from .output.types import FieldSpec, PaginatedResult
from .session import api_session

MAX_PAGE_SIZE: Final = 100

T = TypeVar('T')


async def execute_paginated_query(
    root_field: str,
    variables: Dict[str, Tuple[Any, str]],
    fields: Sequence[FieldSpec],
    *,
    limit: int,
    offset: int,
) -> PaginatedResult:
    if limit > MAX_PAGE_SIZE:
        raise ValueError(f"The page size cannot exceed {MAX_PAGE_SIZE}")
    query = '''
    query($limit:Int!, $offset:Int!, $var_decls) {
      $root_field(
          limit:$limit, offset:$offset, $var_args) {
        items { $fields }
        total_count
      }
    }'''
    query = query.replace('$root_field', root_field)
    query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
    query = query.replace(
        '$var_decls',
        ', '.join(f'${key}: {value[1]}'
                  for key, value in variables.items()),
    )
    query = query.replace(
        '$var_args',
        ', '.join(f'{key}:${key}'
                  for key in variables.keys()),
    )
    query = textwrap.dedent(query).strip()
    var_values = {key: value[0] for key, value in variables.items()}
    var_values['limit'] = limit
    var_values['offset'] = offset
    data = await api_session.get().Admin._query(query, var_values)
    return PaginatedResult(
        total_count=data[root_field]['total_count'],
        items=data[root_field]['items'],
        fields=fields,
    )


async def generate_paginated_results(
    root_field: str,
    variables: Dict[str, Tuple[Any, str]],
    fields: Sequence[FieldSpec],
    *,
    page_offset: int,
    page_size: int,
) -> PaginatedResult:
    if page_size > MAX_PAGE_SIZE:
        raise ValueError(f"The page size cannot exceed {MAX_PAGE_SIZE}")
    if api_session.get().api_version < (6, '20210815'):
        if variables['filter'][0] is not None or variables['order'][0] is not None:
            raise BackendAPIVersionError(
                "filter and order arguments for paginated lists require v6.20210815 or later.",
            )
        # should remove to work with older managers
        variables.pop('filter')
        variables.pop('order')
    offset = page_offset
    while True:
        limit = page_size
        result = await execute_paginated_query(
            root_field, variables, fields,
            limit=limit, offset=offset,
        )
        offset += page_size
        return result



import warnings

import click

from ai.backend.cli.extensions import ExtendedCommandGroup
from ai.backend.client import __version__
from ai.backend.client.cli.types import CLIContext, OutputMode
from ai.backend.client.config import APIConfig, set_config
from ai.backend.client.output import get_output_handler


@click.group(
    cls=ExtendedCommandGroup,
    context_settings={
        'help_option_names': ['-h', '--help'],
    },
)
@click.option('--skip-sslcert-validation',
              help='Skip SSL certificate validation for all API requests.',
              is_flag=True)
@click.option('--output', type=click.Choice(['json', 'console']), default='console',
              help='Set the output style of the command results.')
@click.version_option(version=__version__)
@click.pass_context
def main(ctx: click.Context, skip_sslcert_validation: bool, output: str) -> None:
    """
    Backend.AI command line interface.
    """
    from .announcement import announce
    config = APIConfig(
        skip_sslcert_validation=skip_sslcert_validation,
        announcement_handler=announce,
    )
    set_config(config)

    output_mode = OutputMode(output)
    cli_ctx = CLIContext(
        api_config=config,
        output_mode=output_mode,
    )
    cli_ctx.output = get_output_handler(cli_ctx, output_mode)
    ctx.obj = cli_ctx

    from .pretty import show_warning
    warnings.showwarning = show_warning



import warnings
from functools import update_wrapper
from typing import Any, Callable, Mapping, TypeVar, cast

from click import get_current_context

from ai.backend.cli.types import CliContextInfo
from ai.backend.client.cli.types import CLIContext, OutputMode
from ai.backend.client.config import APIConfig, set_config
from ai.backend.client.output import get_output_handler


def set_client_config(info: Mapping) -> CLIContext:
    from .announcement import announce
    skip_sslcert_validation = info.get('skip_sslcert_validation', False)
    output = info.get('output', 'console')
    config = APIConfig(
        skip_sslcert_validation=skip_sslcert_validation,
        announcement_handler=announce,
    )
    set_config(config)

    output_mode = OutputMode(output)
    cli_ctx = CLIContext(
        api_config=config,
        output_mode=output_mode,
    )
    cli_ctx.output = get_output_handler(cli_ctx, output_mode)

    from .pretty import show_warning
    warnings.showwarning = show_warning
    return cli_ctx


F = TypeVar("F", bound=Callable[..., Any])


def pass_ctx_obj(f: F) -> F:
    def new_func(*args, **kwargs):
        obj = get_current_context().obj
        match obj:
            case CLIContext():
                inner = f(obj, *args, **kwargs)
            case CliContextInfo(info=info):
                inner = f(set_client_config(info), *args, **kwargs)
            case _:
                raise RuntimeError('Invalid Context from client command')
        return inner

    return update_wrapper(cast(F, new_func), f)



import textwrap
from typing import Iterable, Sequence

from ai.backend.client.output.fields import group_fields
from ai.backend.client.output.types import FieldSpec

from ..session import api_session
from .base import BaseFunction, api_function

__all__ = (
    'Group',
)

_default_list_fields = (
    group_fields['id'],
    group_fields['name'],
    group_fields['is_active'],
    group_fields['created_at'],
    group_fields['integration_id'],
)
_default_detail_fields = (
    group_fields['id'],
    group_fields['name'],
    group_fields['description'],
    group_fields['is_active'],
    group_fields['created_at'],
    group_fields['domain_name'],
    group_fields['total_resource_slots'],
    group_fields['allowed_vfolder_hosts'],
    group_fields['integration_id'],
)


class Group(BaseFunction):
    """
    Provides a shortcut of :func:`Group.query()
    <ai.backend.client.admin.Admin.query>` that fetches various group information.

    .. note::

      All methods in this function class require your API access key to
      have the *admin* privilege.
    """

    @api_function
    @classmethod
    async def from_name(
        cls,
        name: str,
        *,
        fields: Iterable[str] = None,
        domain_name: str = None,
    ) -> Sequence[dict]:
        """
        Find the group(s) by its name.
        It may return multiple groups when there are groups with the same name
        in different domains and it is invoked with a super-admin account
        without setting the domain name.

        :param domain_name: Name of domain to get groups from.
        :param fields: Per-group query fields to fetch.
        """
        if fields is None:
            fields = _default_detail_fields
        query = textwrap.dedent("""\
            query($name: String!, $domain_name: String) {
                groups_by_name(name: $name, domain_name: $domain_name) {$fields}
            }
        """)
        query = query.replace('$fields', ' '.join(fields))
        variables = {
            'name': name,
            'domain_name': domain_name,
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['groups_by_name']

    @api_function
    @classmethod
    async def list(
        cls,
        domain_name: str,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        Fetches the list of groups.

        :param domain_name: Name of domain to list groups.
        :param fields: Per-group query fields to fetch.
        """
        if fields is None:
            fields = _default_list_fields
        query = textwrap.dedent("""\
            query($domain_name: String) {
                groups(domain_name: $domain_name) {$fields}
            }
        """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {'domain_name': domain_name}
        data = await api_session.get().Admin._query(query, variables)
        return data['groups']

    @api_function
    @classmethod
    async def detail(
        cls,
        gid: str,
        fields: Sequence[FieldSpec] = _default_detail_fields,
    ) -> dict:
        """
        Fetch information of a group with group ID.

        :param gid: ID of the group to fetch.
        :param fields: Additional per-group query fields to fetch.
        """
        if fields is None:
            fields = _default_detail_fields
        query = textwrap.dedent("""\
            query($gid: UUID!) {
                group(id: $gid) {$fields}
            }
        """)
        query = query.replace('$fields', ' '.join(f.field_ref for f in fields))
        variables = {'gid': gid}
        data = await api_session.get().Admin._query(query, variables)
        return data['group']

    @api_function
    @classmethod
    async def create(cls, domain_name: str, name: str, description: str = '',
                     is_active: bool = True, total_resource_slots: str = None,
                     allowed_vfolder_hosts: Iterable[str] = None,
                     integration_id: str = None,
                     fields: Iterable[str] = None) -> dict:
        """
        Creates a new group with the given options.
        You need an admin privilege for this operation.
        """
        if fields is None:
            fields = ('id', 'domain_name', 'name')
        query = textwrap.dedent("""\
            mutation($name: String!, $input: GroupInput!) {
                create_group(name: $name, props: $input) {
                    ok msg group {$fields}
                }
            }
        """)
        query = query.replace('$fields', ' '.join(fields))
        variables = {
            'name': name,
            'input': {
                'description': description,
                'is_active': is_active,
                'domain_name': domain_name,
                'total_resource_slots': total_resource_slots,
                'allowed_vfolder_hosts': allowed_vfolder_hosts,
                'integration_id': integration_id,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['create_group']

    @api_function
    @classmethod
    async def update(cls, gid: str, name: str = None, description: str = None,
                     is_active: bool = None, total_resource_slots: str = None,
                     allowed_vfolder_hosts: Iterable[str] = None,
                     integration_id: str = None,
                     fields: Iterable[str] = None) -> dict:
        """
        Update existing group.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent("""\
            mutation($gid: UUID!, $input: ModifyGroupInput!) {
                modify_group(gid: $gid, props: $input) {
                    ok msg
                }
            }
        """)
        variables = {
            'gid': gid,
            'input': {
                'name': name,
                'description': description,
                'is_active': is_active,
                'total_resource_slots': total_resource_slots,
                'allowed_vfolder_hosts': allowed_vfolder_hosts,
                'integration_id': integration_id,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['modify_group']

    @api_function
    @classmethod
    async def delete(cls, gid: str):
        """
        Inactivates the existing group. Does not actually delete it for safety.
        """
        query = textwrap.dedent("""\
            mutation($gid: UUID!) {
                delete_group(gid: $gid) {
                    ok msg
                }
            }
        """)
        variables = {'gid': gid}
        data = await api_session.get().Admin._query(query, variables)
        return data['delete_group']

    @api_function
    @classmethod
    async def purge(cls, gid: str):
        """
        Delete the existing group. This action cannot be undone.
        """
        query = textwrap.dedent("""\
            mutation($gid: UUID!) {
                purge_group(gid: $gid) {
                    ok msg
                }
            }
        """)
        variables = {'gid': gid}
        data = await api_session.get().Admin._query(query, variables)
        return data['purge_group']

    @api_function
    @classmethod
    async def add_users(cls, gid: str, user_uuids: Iterable[str],
                        fields: Iterable[str] = None) -> dict:
        """
        Add users to a group.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent("""\
            mutation($gid: UUID!, $input: ModifyGroupInput!) {
                modify_group(gid: $gid, props: $input) {
                    ok msg
                }
            }
        """)
        variables = {
            'gid': gid,
            'input': {
                'user_update_mode': 'add',
                'user_uuids': user_uuids,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['modify_group']

    @api_function
    @classmethod
    async def remove_users(cls, gid: str, user_uuids: Iterable[str],
                           fields: Iterable[str] = None) -> dict:
        """
        Remove users from a group.
        You need an admin privilege for this operation.
        """
        query = textwrap.dedent("""\
            mutation($gid: UUID!, $input: ModifyGroupInput!) {
                modify_group(gid: $gid, props: $input) {
                    ok msg
                }
            }
        """)
        variables = {
            'gid': gid,
            'input': {
                'user_update_mode': 'remove',
                'user_uuids': user_uuids,
            },
        }
        data = await api_session.get().Admin._query(query, variables)
        return data['modify_group']



import asyncio
from pathlib import Path
from typing import Mapping, Optional, Sequence, Union

import aiohttp
import janus
from aiotusclient import client
from tqdm import tqdm
from yarl import URL

from ai.backend.client.output.fields import vfolder_fields
from ai.backend.client.output.types import FieldSpec, PaginatedResult

from ..compat import current_loop
from ..config import DEFAULT_CHUNK_SIZE, MAX_INFLIGHT_CHUNKS
from ..exceptions import BackendClientError
from ..pagination import generate_paginated_results
from ..request import Request
from .base import BaseFunction, api_function

__all__ = (
    'VFolder',
)

_default_list_fields = (
    vfolder_fields['host'],
    vfolder_fields['name'],
    vfolder_fields['created_at'],
    vfolder_fields['creator'],
    vfolder_fields['group_id'],
    vfolder_fields['permission'],
    vfolder_fields['ownership_type'],
)


class VFolder(BaseFunction):

    def __init__(self, name: str):
        self.name = name

    @api_function
    @classmethod
    async def create(
        cls,
        name: str,
        host: str = None,
        unmanaged_path: str = None,
        group: str = None,
        usage_mode: str = 'general',
        permission: str = 'rw',
        quota: str = '0',
        cloneable: bool = False,
    ):
        rqst = Request('POST', '/folders')
        rqst.set_json({
            'name': name,
            'host': host,
            'unmanaged_path': unmanaged_path,
            'group': group,
            'usage_mode': usage_mode,
            'permission': permission,
            'quota': quota,
            'cloneable': cloneable,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def delete_by_id(cls, oid):
        rqst = Request('DELETE', '/folders')
        rqst.set_json({'id': oid})
        async with rqst.fetch():
            return {}

    @api_function
    @classmethod
    async def list(cls, list_all=False):
        rqst = Request('GET', '/folders')
        rqst.set_json({'all': list_all})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        group: str = None,
        *,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str = None,
        order: str = None,
    ) -> PaginatedResult[dict]:
        """
        Fetches the list of vfolders. Domain admins can only get domain vfolders.

        :param group: Fetch vfolders in a specific group.
        :param fields: Additional per-vfolder query fields to fetch.
        """
        return await generate_paginated_results(
            'vfolder_list',
            {
                'group_id': (group, 'UUID'),
                'filter': (filter, 'String'),
                'order': (order, 'String'),
            },
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    @classmethod
    async def list_hosts(cls):
        rqst = Request('GET', '/folders/_/hosts')
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_all_hosts(cls):
        rqst = Request('GET', '/folders/_/all_hosts')
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_allowed_types(cls):
        rqst = Request('GET', '/folders/_/allowed_types')
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def info(self):
        rqst = Request('GET', '/folders/{0}'.format(self.name))
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete(self):
        rqst = Request('DELETE', '/folders/{0}'.format(self.name))
        async with rqst.fetch():
            return {}

    @api_function
    async def rename(self, new_name):
        rqst = Request('POST', '/folders/{0}/rename'.format(self.name))
        rqst.set_json({
            'new_name': new_name,
        })
        async with rqst.fetch() as resp:
            self.name = new_name
            return await resp.text()

    @api_function
    async def download(
        self,
        relative_paths: Sequence[Union[str, Path]],
        *,
        basedir: Union[str, Path] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        show_progress: bool = False,
        address_map: Optional[Mapping[str, str]] = None,
    ) -> None:
        base_path = (Path.cwd() if basedir is None else Path(basedir).resolve())
        for relpath in relative_paths:
            file_path = base_path / relpath
            rqst = Request('POST',
                           '/folders/{}/request-download'.format(self.name))
            rqst.set_json({
                'path': str(relpath),
            })
            async with rqst.fetch() as resp:
                download_info = await resp.json()
                overriden_url = download_info['url']
                if address_map:
                    if download_info['url'] in address_map:
                        overriden_url = address_map[download_info['url']]
                    else:
                        raise BackendClientError(
                            'Overriding storage proxy addresses are given, '
                            'but no url matches with any of them.\n',
                        )

                download_url = URL(overriden_url).with_query({
                    'token': download_info['token'],
                })

            def _write_file(file_path: Path, q: janus._SyncQueueProxy[bytes]):
                with open(file_path, 'wb') as f:
                    while True:
                        chunk = q.get()
                        if not chunk:
                            return
                        f.write(chunk)
                        q.task_done()

            if show_progress:
                print(f"Downloading to {file_path} ...")
            async with aiohttp.ClientSession() as client:
                # TODO: ranged requests to continue interrupted downloads with automatic retries
                async with client.get(download_url, ssl=False) as raw_resp:
                    size = int(raw_resp.headers['Content-Length'])
                    if file_path.exists():
                        raise RuntimeError('The target file already exists', file_path.name)
                    q: janus.Queue[bytes] = janus.Queue(MAX_INFLIGHT_CHUNKS)
                    try:
                        with tqdm(
                            total=size,
                            unit='bytes',
                            unit_scale=True,
                            unit_divisor=1024,
                            disable=not show_progress,
                        ) as pbar:
                            loop = current_loop()
                            writer_fut = loop.run_in_executor(None, _write_file, file_path, q.sync_q)
                            await asyncio.sleep(0)
                            while True:
                                chunk = await raw_resp.content.read(chunk_size)
                                pbar.update(len(chunk))
                                if not chunk:
                                    break
                                await q.async_q.put(chunk)
                    finally:
                        await q.async_q.put(b'')
                        await writer_fut
                        q.close()
                        await q.wait_closed()

    @api_function
    async def upload(
        self,
        files: Sequence[Union[str, Path]],
        *,
        basedir: Union[str, Path] = None,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        address_map: Optional[Mapping[str, str]] = None,
        show_progress: bool = False,
    ) -> None:
        base_path = (Path.cwd() if basedir is None else Path(basedir).resolve())
        if basedir:
            files = [basedir / Path(file) for file in files]
        else:
            files = [Path(file).resolve() for file in files]
        for file_path in files:
            file_size = Path(file_path).stat().st_size
            rqst = Request('POST',
                           '/folders/{}/request-upload'.format(self.name))
            rqst.set_json({
                'path': "{}".format(str(Path(file_path).relative_to(base_path))),
                'size': int(file_size),
            })
            async with rqst.fetch() as resp:
                upload_info = await resp.json()
                overriden_url = upload_info['url']
                if address_map:
                    if upload_info['url'] in address_map:
                        overriden_url = address_map[upload_info['url']]
                    else:
                        raise BackendClientError(
                            'Overriding storage proxy addresses are given, '
                            'but no url matches with any of them.\n',
                        )
                upload_url = URL(overriden_url).with_query({
                    'token': upload_info['token'],
                })
            tus_client = client.TusClient()
            if basedir:
                input_file = open(base_path / file_path, "rb")
            else:
                input_file = open(str(Path(file_path).relative_to(base_path)), "rb")
            print(f"Uploading {base_path / file_path} via {upload_info['url']} ...")
            # TODO: refactor out the progress bar
            uploader = tus_client.async_uploader(
                file_stream=input_file,
                url=upload_url,
                upload_checksum=False,
                chunk_size=chunk_size,
            )
            return await uploader.upload()

    @api_function
    async def mkdir(
        self,
        path: Union[str, Path],
        parents: Optional[bool] = False,
        exist_ok: Optional[bool] = False,
    ):
        rqst = Request('POST',
                       '/folders/{}/mkdir'.format(self.name))
        rqst.set_json({
            'path': path,
            'parents': parents,
            'exist_ok': exist_ok,
        })
        async with rqst.fetch() as resp:
            return await resp.text()

    @api_function
    async def rename_file(self, target_path: str, new_name: str):
        rqst = Request('POST',
                       '/folders/{}/rename-file'.format(self.name))
        rqst.set_json({
            'target_path': target_path,
            'new_name': new_name,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def move_file(self, src_path: str, dst_path: str):
        rqst = Request('POST',
                       '/folders/{}/move-file'.format(self.name))
        rqst.set_json({
            'src': src_path,
            'dst': dst_path,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def delete_files(self,
                           files: Sequence[Union[str, Path]],
                           recursive: bool = False):
        rqst = Request('DELETE',
                       '/folders/{}/delete-files'.format(self.name))
        rqst.set_json({
            'files': files,
            'recursive': recursive,
        })
        async with rqst.fetch() as resp:
            return await resp.text()

    @api_function
    async def list_files(self, path: Union[str, Path] = '.'):
        rqst = Request('GET', '/folders/{}/files'.format(self.name))
        rqst.set_json({
            'path': path,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def invite(self, perm: str, emails: Sequence[str]):
        rqst = Request('POST', '/folders/{}/invite'.format(self.name))
        rqst.set_json({
            'perm': perm, 'user_ids': emails,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def invitations(cls):
        rqst = Request('GET', '/folders/invitations/list')
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def accept_invitation(cls, inv_id: str):
        rqst = Request('POST', '/folders/invitations/accept')
        rqst.set_json({'inv_id': inv_id})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def delete_invitation(cls, inv_id: str):
        rqst = Request('DELETE', '/folders/invitations/delete')
        rqst.set_json({'inv_id': inv_id})
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def get_fstab_contents(cls, agent_id=None):
        rqst = Request('GET', '/folders/_/fstab')
        rqst.set_json({
            'agent_id': agent_id,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def get_performance_metric(cls, folder_host: str):
        rqst = Request('GET', '/folders/_/perf-metric')
        rqst.set_json({
            'folder_host': folder_host,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def list_mounts(cls):
        rqst = Request('GET', '/folders/_/mounts')
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def mount_host(cls, name: str, fs_location: str, options=None,
                         edit_fstab: bool = False):
        rqst = Request('POST', '/folders/_/mounts')
        rqst.set_json({
            'name': name,
            'fs_location': fs_location,
            'options': options,
            'edit_fstab': edit_fstab,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def umount_host(cls, name: str, edit_fstab: bool = False):
        rqst = Request('DELETE', '/folders/_/mounts')
        rqst.set_json({
            'name': name,
            'edit_fstab': edit_fstab,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def share(self, perm: str, emails: Sequence[str]):
        rqst = Request('POST', '/folders/{}/share'.format(self.name))
        rqst.set_json({
            'permission': perm, 'emails': emails,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def unshare(self, emails: Sequence[str]):
        rqst = Request('DELETE', '/folders/{}/unshare'.format(self.name))
        rqst.set_json({
            'emails': emails,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def leave(self):
        rqst = Request('POST', '/folders/{}/leave'.format(self.name))
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def clone(self, target_name: str, target_host: str = None,
                    usage_mode: str = 'general', permission: str = 'rw'):
        rqst = Request('POST', '/folders/{}/clone'.format(self.name))
        rqst.set_json({
            'target_name': target_name,
            'target_host': target_host,
            'usage_mode': usage_mode,
            'permission': permission,
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    async def update_options(self, name: str, permission: str = None,
                             cloneable: bool = None):
        rqst = Request('POST', '/folders/{}/update-options'.format(self.name))
        rqst.set_json({
            'cloneable': cloneable,
            'permission': permission,
        })
        async with rqst.fetch() as resp:
            return await resp.text()



from __future__ import annotations

from .formatters import (
    AgentStatFormatter,
    ContainerListFormatter,
    DependencyListFormatter,
    GroupListFormatter,
    KernelStatFormatter,
    SubFieldOutputFormatter,
    mibytes_output_formatter,
    nested_dict_formatter,
    resource_slot_formatter,
    sizebytes_output_formatter,
)
from .types import FieldSet, FieldSpec

container_fields = FieldSet([
    FieldSpec('id', "Kernel ID", alt_name='kernel_id'),
    FieldSpec('cluster_role'),
    FieldSpec('cluster_idx'),
    FieldSpec('cluster_hostname'),
    FieldSpec('session_id', "Session ID"),
    FieldSpec('image'),
    FieldSpec('registry'),
    FieldSpec('status'),
    FieldSpec('status_info'),
    FieldSpec('status_data', formatter=nested_dict_formatter),
    FieldSpec('status_changed'),
    FieldSpec('agent'),
    FieldSpec('container_id'),
    FieldSpec('resource_opts', formatter=nested_dict_formatter),
    FieldSpec('occupied_slots', formatter=resource_slot_formatter),
    FieldSpec('live_stat', formatter=KernelStatFormatter()),
    FieldSpec('last_stat', formatter=KernelStatFormatter()),
])


agent_fields = FieldSet([
    FieldSpec('id'),
    FieldSpec('status'),
    FieldSpec('status_changed'),
    FieldSpec('region'),
    FieldSpec('architecture'),
    FieldSpec('scaling_group'),
    FieldSpec('schedulable'),
    FieldSpec('available_slots', formatter=resource_slot_formatter),
    FieldSpec('occupied_slots', formatter=resource_slot_formatter),
    FieldSpec('addr'),
    FieldSpec('first_contact'),
    FieldSpec('lost_at'),
    FieldSpec('live_stat', formatter=AgentStatFormatter()),
    FieldSpec('version'),
    FieldSpec('compute_plugins'),
    FieldSpec('hardware_metadata', formatter=nested_dict_formatter),
    FieldSpec('compute_containers', subfields=container_fields,
              formatter=ContainerListFormatter()),
    # legacy fields
    FieldSpec('cpu_cur_pct', 'CPU Usage (%)'),
    FieldSpec('mem_cur_bytes', 'Used Memory (MiB)', formatter=mibytes_output_formatter),
])

domain_fields = FieldSet([
    FieldSpec('name'),
    FieldSpec('description'),
    FieldSpec('is_active'),
    FieldSpec('created_at'),
    FieldSpec('total_resource_slots', formatter=resource_slot_formatter),
    FieldSpec('allowed_vfolder_hosts'),
    FieldSpec('allowed_docker_registries'),
    FieldSpec('integration_id'),
])

group_fields = FieldSet([
    FieldSpec('id'),
    FieldSpec('name'),
    FieldSpec('description'),
    FieldSpec('is_active'),
    FieldSpec('created_at'),
    FieldSpec('domain_name'),
    FieldSpec('total_resource_slots', formatter=resource_slot_formatter),
    FieldSpec('allowed_vfolder_hosts'),
    FieldSpec('integration_id'),
])


image_fields = FieldSet([
    FieldSpec('name'),
    FieldSpec('registry'),
    FieldSpec('architecture'),
    FieldSpec('tag'),
    FieldSpec('digest'),
    FieldSpec('size_bytes', formatter=sizebytes_output_formatter),
    FieldSpec('aliases'),
])


keypair_fields = FieldSet([
    FieldSpec('user_id', "Email"),
    FieldSpec('user_info { full_name }', "Full Name", alt_name='full_name',
              formatter=SubFieldOutputFormatter("full_name")),
    FieldSpec('access_key'),
    FieldSpec('secret_key'),
    FieldSpec('is_active'),
    FieldSpec('is_admin'),
    FieldSpec('created_at'),
    FieldSpec('modified_at'),
    FieldSpec('last_used'),
    FieldSpec('resource_policy'),
    FieldSpec('rate_limit'),
    FieldSpec('concurrency_used'),
    FieldSpec('ssh_public_key'),
    FieldSpec('ssh_private_key'),
    FieldSpec('dotfiles'),
    FieldSpec('bootstrap_script'),
])


keypair_resource_policy_fields = FieldSet([
    FieldSpec('name'),
    FieldSpec('created_at'),
    FieldSpec('total_resource_slots'),
    FieldSpec('max_concurrent_sessions'),  # formerly concurrency_limit
    FieldSpec('max_vfolder_count'),
    FieldSpec('max_vfolder_size', formatter=sizebytes_output_formatter),
    FieldSpec('idle_timeout'),
    FieldSpec('max_containers_per_session'),
    FieldSpec('allowed_vfolder_hosts'),
])


scaling_group_fields = FieldSet([
    FieldSpec('name'),
    FieldSpec('description'),
    FieldSpec('is_active'),
    FieldSpec('created_at'),
    FieldSpec('driver'),
    FieldSpec('driver_opts', formatter=nested_dict_formatter),
    FieldSpec('scheduler'),
    FieldSpec('scheduler_opts', formatter=nested_dict_formatter),
])


session_fields = FieldSet([
    FieldSpec('id', "Kernel ID", alt_name='kernel_id'),
    FieldSpec('tag'),
    FieldSpec('name'),
    FieldSpec('type'),
    FieldSpec('session_id', "Session ID"),
    FieldSpec('image'),
    FieldSpec('registry'),
    FieldSpec('cluster_template'),
    FieldSpec('cluster_mode'),
    FieldSpec('cluster_size'),
    FieldSpec('domain_name'),
    FieldSpec('group_name', "Project/Group"),
    FieldSpec('group_id'),
    FieldSpec('user_email'),
    FieldSpec('user_id'),
    FieldSpec('access_key', "Owner Access Key"),
    FieldSpec('created_user_email'),
    FieldSpec('created_user_id'),
    FieldSpec('status'),
    FieldSpec('status_info'),
    FieldSpec('status_data', formatter=nested_dict_formatter),
    FieldSpec('status_changed', "Last Updated"),
    FieldSpec('created_at'),
    FieldSpec('terminated_at'),
    FieldSpec('starts_at'),
    FieldSpec('startup_command'),
    FieldSpec('result'),
    FieldSpec('resoucre_opts', formatter=nested_dict_formatter),
    FieldSpec('scaling_group'),
    FieldSpec('service_ports', formatter=nested_dict_formatter),
    FieldSpec('mounts'),
    FieldSpec('occupied_slots', formatter=resource_slot_formatter),
    FieldSpec(
        'containers',
        subfields=container_fields,
        formatter=ContainerListFormatter(),
    ),
    FieldSpec(
        'dependencies { name id }',
        formatter=DependencyListFormatter(),
    ),
])

session_fields_v5 = FieldSet([
    FieldSpec(
        'containers',
        subfields=FieldSet([
            FieldSpec('id', "Kernel ID", alt_name='kernel_id'),
            FieldSpec('session_id', "Session ID"),
            FieldSpec('role'),
            FieldSpec('agent'),
            FieldSpec('image'),
            FieldSpec('status'),
            FieldSpec('status_info'),
            FieldSpec('status_data', formatter=nested_dict_formatter),
            FieldSpec('status_changed'),
            FieldSpec('occupied_slots', formatter=resource_slot_formatter),
            FieldSpec('live_stat', formatter=KernelStatFormatter()),
            FieldSpec('last_stat', formatter=KernelStatFormatter()),
        ]),
        formatter=ContainerListFormatter(),
    ),
])


storage_fields = FieldSet([
    FieldSpec('id'),
    FieldSpec('backend'),
    FieldSpec('fsprefix'),
    FieldSpec('path'),
    FieldSpec('capabilities'),
    FieldSpec('hardware_metadata', formatter=nested_dict_formatter),
    FieldSpec('performance_metric', formatter=nested_dict_formatter),
    FieldSpec('usage', formatter=nested_dict_formatter),
])


user_fields = FieldSet([
    FieldSpec('uuid'),
    FieldSpec('username'),
    FieldSpec('email'),
    # password is not queriable!
    FieldSpec('need_password_change'),
    FieldSpec('full_name'),
    FieldSpec('description'),
    FieldSpec('is_active'),
    FieldSpec('status'),
    FieldSpec('status_info'),
    FieldSpec('created_at'),
    FieldSpec('modified_at'),
    FieldSpec('domain_name'),
    FieldSpec('role'),
    FieldSpec('groups { id name }', formatter=GroupListFormatter()),
])


vfolder_fields = FieldSet([
    FieldSpec('id'),
    FieldSpec('host'),
    FieldSpec('name'),
    FieldSpec('user', alt_name='user_id'),
    FieldSpec('group', alt_name='group_id'),
    FieldSpec('creator'),
    FieldSpec('unmanaged_path'),
    FieldSpec('usage_mode'),
    FieldSpec('permission'),
    FieldSpec('ownership_type'),
    FieldSpec('max_files'),
    FieldSpec('max_size'),
    FieldSpec('created_at'),
    FieldSpec('last_used'),
    FieldSpec('num_files'),
    FieldSpec('cur_size'),
    FieldSpec('cloneable'),
])



import shutil
import sys
from typing import Any, Callable, Iterator, List, Literal, MutableMapping, Sequence

import click
from tabulate import tabulate

from ai.backend.client.output.types import FieldSpec

from ..pagination import MAX_PAGE_SIZE


def get_preferred_page_size() -> int:
    return min(MAX_PAGE_SIZE, shutil.get_terminal_size((80, 20)).lines)


_Item = MutableMapping[str, Any]


def tabulate_items(
    items: Iterator[_Item],
    fields: Sequence[FieldSpec],
    *,
    page_size: int = None,
    item_formatter: Callable[[_Item], None] = None,
    tablefmt: Literal['simple', 'plain', 'github'] = 'simple',
) -> Iterator[str]:
    is_first = True
    output_count = 0
    buffered_items: List[_Item] = []

    # check table header/footer sizes
    header_height = 0
    if tablefmt in ('simple', 'github'):
        header_height = 2
    assert header_height >= 0

    def _tabulate_buffer() -> Iterator[str]:
        table = tabulate(
            [
                [
                    f.formatter.format_console(v, f) for f, v in zip(fields, item.values())
                ] for item in buffered_items
            ],
            headers=(
                [] if tablefmt == 'plain'
                else [field.humanized_name for field in fields]
            ),
            tablefmt=tablefmt,
        )
        table_rows = table.splitlines()
        if is_first:
            yield from (row + '\n' for row in table_rows)
        else:
            # strip the header for continued page outputs
            yield from (row + '\n' for row in table_rows[header_height:])

    # If we iterate until the end of items, pausing the terminal output
    # would not have effects for avoiding unnecessary queries for subsequent pages.
    # Let's buffer the items and split the formatting per page.
    if page_size is None:
        table_height = shutil.get_terminal_size((80, 20)).lines
    else:
        table_height = page_size
    page_size = max(table_height - header_height - 1, 10)
    for item in items:
        if item_formatter is not None:
            item_formatter(item)
        buffered_items.append(item)
        output_count += 1
        if output_count == page_size:
            yield from _tabulate_buffer()
            buffered_items.clear()
            is_first = False
            output_count = 0
            page_size = max(table_height - 1, 10)
    if output_count > 0:
        yield from _tabulate_buffer()


def echo_via_pager(
    text_generator: Iterator[str],
    break_callback: Callable[[], None] = None,
) -> None:
    """
    A variant of ``click.echo_via_pager()`` which implements our own simplified pagination.
    The key difference is that it holds the generator for each page, so that the generator
    won't continue querying the next results unless continued, avoiding server overloads.
    """
    # TODO: support PageUp & PageDn by buffering the output
    terminal_height = shutil.get_terminal_size((80, 20)).lines
    line_count = 0
    for text in text_generator:
        line_count += text.count('\n')
        click.echo(text, nl=False)
        if line_count == terminal_height - 1:
            if sys.stdin.isatty() and sys.stdout.isatty():
                click.echo(':', nl=False)
                # Pause the terminal so that we don't execute next-page queries indefinitely.
                # Since click.pause() ignores KeyboardInterrupt, we just use click.getchar()
                # to allow user interruption.
                k = click.getchar(echo=False)
                if k in ('q', 'Q'):
                    if break_callback is not None:
                        break_callback()
                    break
                click.echo('\r', nl=False)
            line_count = 0
